# -*- coding: utf-8 -*-
import os

working_settings = {}

DEFAULTS = {
    'PROJECT_NAME': os.environ.get('PROJECT_NAME', ''),
    'PROJECT_VERSION': os.environ.get('PROJECT_VERSION', ''),
    'HTTP_PORT': 80,
    'WORKERS': 1,
    'INSPECTOR_REPORT_INTERVAL': 30,

    'LOG': {
        'version': 1,
        'disable_existing_loggers': False,
        'loggers': {
            "": {
                "level": 'WARNING',
                "handlers": ["default"],
                'propagate': False,
            },
            'wings_sanic': {
                "level": 'INFO',
                "handlers": ["wings_sanic"],
                'propagate': False,
            }
        },
        'handlers': {
            "default": {
                "class": "logging.StreamHandler",
                "formatter": "json",
                "level": 'WARNING',
            },
            "wings_sanic": {
                "class": "logging.StreamHandler",
                "formatter": "json",
                "level": 'INFO',
            }
        },
        'formatters': {
            "json": {
                "class": "wings_sanic.log_formatter.JsonFormatter"
            }
        },
    },

    # default LOG(JsonFormatter) is useful for elk, but not friendly for develop debug.
    # for convenience for develop, set True, then use sanic's log and its access_log is available
    'DEV': False,
    'CORS': False,
    'DEBUG': False,

    'BLUEPRINTS': [],
    'DEFAULT_CONTEXT': {
        'response_shape': 'wings_sanic.views.ResponseShape',
        'serialize_when_none': True
    },
    'EXCEPTION_HANDLER': 'wings_sanic.views.exception_handler',

    'SWAGGER': {
        'info': {
            "version": os.environ.get('PROJECT_VERSION', '1.0.0'),
            "title": os.environ.get('PROJECT_NAME', 'API'),
            "description": '',
            "termsOfService": None,
            "contact": {
                "email": None
            },
            "license": {
                "email": None,
                "url": None
            }
        },
        'schemes': ['http']
    },
    'MQ_SERVERS': {
        # 'default': {
        #     'server': 'wings_sanic.events.rabbitmq_server.MqServer',
        #     'url': 'amqp://guest:guest@127.0.0.1:5672',
        #     'exchange': 'test',
        #     'exchange_type': 'topic',
        #     'reconnect_delay': 5.0
        # }
    },
    'EVENT_HANDLE_TIMEOUT': 10,
    'EVENT_MAX_RETRY': -1,

    # rpc， mq传递上下文信息时忽略的key
    'IGNORE_CONTEXT_WHEN_DELIVERY':[
        'messages',
        'response_shape',
        'serialize_when_none'
    ]
}


def load(**user_settings):
    working_settings.update(DEFAULTS)
    working_settings.update(user_settings)


def get(attr):
    try:
        # Check if present in user settings
        return working_settings[attr]
    except KeyError:
        try:
            return DEFAULTS[attr]
        except KeyError:
            raise AttributeError("Invalid setting: '%s'" % attr)



import inspect
import os
import re
from itertools import repeat

from sanic.blueprints import Blueprint
from sanic.response import json
from sanic.views import CompositionView

from wings_sanic import settings, utils, serializers
from wings_sanic.views import get_response_shape

blueprint = Blueprint('swagger', url_prefix='swagger')

_spec = {}


def __summary_description(doc_string):
    doc_string = (doc_string or "").strip()
    s = doc_string.split('\n', 1)
    if len(s) == 1:
        s.append("")
    return s[0].strip(), s[1].strip()


@blueprint.listener('before_server_start')
def build_spec(app, loop):
    _spec['swagger'] = '2.0'
    _spec.update(settings.get('SWAGGER'))

    # --------------------------------------------------------------- #
    # Paths
    # --------------------------------------------------------------- #
    paths = {}
    # tags =
    for uri, route in app.router.routes_all.items():
        if uri.startswith("/swagger"):
            continue

        # Build list of methods and their handler functions
        handler_type = type(route.handler)
        if handler_type is CompositionView:
            method_handlers = route.handler.handlers.items()
        else:
            method_handlers = zip(route.methods, repeat(route.handler))

        methods = {}
        for _method, _handler in method_handlers:
            metadata = utils.get_value(_handler, 'metadata')
            if _method == 'OPTIONS' or not metadata or metadata.swagger_exclude:
                continue
            parameters = []
            # header
            for name, field in (utils.get_value(metadata.header_serializer, 'fields') or {}).items():
                parameter = field.openapi_spec()
                parameter.update({'in': 'header'})
                parameters.append(parameter)

            # path
            for name, field in (utils.get_value(metadata.path_serializer, 'fields') or {}).items():
                parameter = field.openapi_spec()
                parameter.update({'in': 'path', 'required': True})
                parameters.append(parameter)

            # query
            for name, field in (utils.get_value(metadata.query_serializer, 'fields') or {}).items():
                parameter = field.openapi_spec()
                parameter.update({'in': 'query'})
                parameters.append(parameter)

            # body
            if metadata.body_serializer:
                spec = metadata.body_serializer.openapi_spec()
                parameters.append({
                    'in': 'body',
                    'name': 'body',
                    'required': True,
                    'schema': spec
                })

            # response
            response_spec = None
            if metadata.response_serializer:
                response_spec = metadata.response_serializer.openapi_spec()
            response_shape = get_response_shape(metadata.context)
            response_spec = response_shape.swagger(response_spec)

            summary, description = __summary_description(inspect.cleandoc(_handler.__doc__ or ""))

            endpoint = {
                'operationId': utils.meth_str(_handler),
                'summary': summary,
                'description': description,
                'consumes': ['application/json'],
                'produces': ['application/json'],
                'tags': metadata.tags,
                'parameters': parameters,
                'responses': {
                    "200": {
                        "description": None,
                        "examples": None,
                        "schema": response_spec
                    }
                },
            }

            methods[_method.lower()] = endpoint

        uri_parsed = uri
        for parameter in route.parameters:
            uri_parsed = re.sub('<' + parameter.name + '.*?>', '{' + parameter.name + '}', uri_parsed)

        paths[uri_parsed] = methods

    # --------------------------------------------------------------- #
    # Definitions
    # --------------------------------------------------------------- #

    _spec['definitions'] = {cls.__name__: definition for cls, (obj, definition) in serializers.definitions.items()}

    _spec['paths'] = paths


@blueprint.route('/openapi.json')
def spec(request, *args, **kwargs):
    return json(_spec)


dir_path = os.path.dirname(os.path.realpath(__file__))
blueprint.static('/', dir_path + '/index.html')
blueprint.static('/', dir_path)



# -*- coding: utf-8 -*-
import asyncio
import contextvars
import json
import logging
import uuid
from functools import wraps

from . import datetime_helper, registry, context_var, settings, utils

logger = logging.getLogger(__name__)


class DomainEvent:
    def __init__(self, event_name, **kwargs):
        if 'version' in kwargs:
            kwargs['version'] = kwargs['version'] + 1
        self.occur_on = datetime_helper.get_utc_time()
        self.event_name = event_name
        self.__dict__.update(kwargs)


def __get_mq_server(server_name):
    server = registry.get(server_name, 'mq_servers')
    if not server:
        raise Exception('cannot find named "{0}" mq_server'.format(server_name))
    return server


async def publish(event: DomainEvent, mq_server='default', send_after_done=True):
    """
    发送事件
    """
    await publish_message(event.event_name, event, mq_server, send_after_done)


async def publish_message(routing_key: str, message, mq_server='default', send_after_done=True):
    """
    发送消息
    """
    ctx_delivery = {}
    for k, v in context_var.get().items():
        if k not in settings.get('IGNORE_CONTEXT_WHEN_DELIVERY'):
            ctx_delivery[k] = v
    body = {
        'message': utils.to_primitive(message),
        'context': ctx_delivery
    }
    body = json.dumps(body)
    server = __get_mq_server(mq_server)
    if send_after_done:
        context_var.get()['messages'].append({'mq_server': server, 'body': body, 'routing_key': routing_key})
    else:
        await server.publish(routing_key, body)


def handler(event_name, mq_server='default', msg_type=DomainEvent, timeout=None, max_retry=None):
    """
    添加事件处理器的装饰器
    :param event_name: 事件名
    :param mq_server: mq_server名字
    :param msg_type: 将接受到的内容转化成该msg_type对应的类型，如果为None或者转化不成功则返回原字符串
    :param timeout: 处理的超时时长，默认值10秒，优先级 默认值 < settings设置 < 装饰器设置
    :param max_retry: 最大重试次数，默认值-1(无限重试)，优先级 默认值 < settings设置 < 装饰器设置
    :return: 
    """
    timeout = timeout if timeout is not None else settings.get('EVENT_HANDLE_TIMEOUT')
    timeout = timeout if timeout is not None else 10
    max_retry = max_retry if max_retry is not None else settings.get('EVENT_MAX_RETRY')
    max_retry = max_retry if max_retry is not None else -1

    def decorator(func):
        @wraps(func)
        async def wrapper(content, retried_count):
            server = __get_mq_server(mq_server)
            data = utils.instance_from_json(content)
            ctx = utils.get_value(data, 'context', {})
            ctx['trace_id'] = utils.get_value(ctx, 'trace_id', str(uuid.uuid4().hex))
            ctx['messages'] = []
            context_var.set(ctx)

            message = utils.get_value(data, 'message', data)
            message = utils.instance_from_json(message, cls=msg_type)
            try:
                if asyncio.iscoroutinefunction(func):
                    fu = asyncio.ensure_future(func(message), loop=server.loop)
                else:
                    context = contextvars.copy_context()
                    fu = server.loop.run_in_executor(None, context.run, func, message)
                await asyncio.wait_for(fu, timeout, loop=server.loop)
                commit_events()
                logger.info(
                    f'handle message success. event_name:{event_name}, handler:{utils.meth_str(func)}, '
                    f'retried_count: {retried_count}')
            except Exception as ex:
                error_info = f"event_name: {event_name}, handler: {utils.meth_str(func)}, " \
                             f"retried_count: {retried_count}, messages: \n{message}"
                logger.error(error_info, exc_info=ex)
                raise ex
            finally:
                context_var.set(None)

        # 将handler注册到registry，app启动后初始化订阅
        handlers_cur = registry.get(mq_server, 'event_handlers') or set()
        handlers_cur.add((event_name, wrapper, max_retry))
        registry.set(mq_server, handlers_cur, 'event_handlers')

        return func

    return decorator


def commit_events():
    try:
        messages = utils.get_value(context_var.get(), 'messages', [])
        for i in messages:
            server = i['mq_server']
            asyncio.run_coroutine_threadsafe(server.publish(i['routing_key'], i['body']), server.loop)
    except Exception as ex:
        logger.error('send event error', exc_info=ex)
